@@ -7,8 +7,8 @@ use strict;
 BEGIN {  # Make a DEBUG constant very first thing...
   if(defined &DEBUG) {
   } elsif(($ENV{'TKPODDEBUG'} || '') =~ m/^(\d+)/) { # untaint
-    eval("sub DEBUG () {$1}");
-    die "WHAT? Couldn't eval-up a DEBUG constant!? $@" if $@;
+    my $debug = $1;
+    *DEBUG = sub () { $debug };
   } else {
     *DEBUG = sub () {0};
   }
@@ -19,12 +19,15 @@ use Config;
 use Tk qw(catch);
 use Tk::Frame;
 use Tk::Pod;
-#XXX del: use Tk::Parse;
 use Tk::Pod::SimpleBridge;
 use Tk::Pod::Cache;
+use Tk::Pod::Util qw(is_in_path is_interactive detect_window_manager start_browser);
+
+use vars qw($VERSION @ISA @POD $IDX
+	    @tempfiles @gv_pids $terminal_fallback_warn_shown);
+
+$VERSION = sprintf("%d.%02d", q$Revision: 5.20 $ =~ /(\d+)\.(\d+)/);
 
-use vars qw($VERSION @ISA @POD $IDX);
-$VERSION = substr(q$Revision: 3.20 $, 10) + 1 . "";
 @ISA = qw(Tk::Frame Tk::Pod::SimpleBridge Tk::Pod::Cache);
 
 BEGIN { DEBUG and warn "Running ", __PACKAGE__, "\n" }
@@ -32,7 +35,7 @@ BEGIN { DEBUG and warn "Running ", __PACKAGE__, "\n" }
 Construct Tk::Widget 'PodText';
 
 BEGIN {
-  @POD = (
+  unshift @POD, (
    @INC,
    $ENV{'PATH'} ?
      grep(-d, split($Config{path_sep}, $ENV{'PATH'}))
@@ -45,12 +48,19 @@ BEGIN {
 use Class::Struct;
 struct '_HistoryEntry' => [
     'file'  => '$',
+    'text'  => '$',
     'index' => '$',
 ];
 sub _HistoryEntry::create {
     my $o = shift->new;
-    $o->file(shift);
-    $o->index(shift);
+    my($what, $index) = @_;
+    if (ref $what eq 'HASH') {
+	$o->file($what->{file});
+	$o->text($what->{text});
+    } else {
+	$o->file($what);
+    }
+    $o->index($index);
     $o;
 }
 
@@ -74,7 +84,7 @@ sub Find
    foreach $prefix ("","pod/","pods/")
     {
      my $suffix;
-     foreach $suffix ("",".pod",".pm")
+     foreach $suffix (".pod",".pm",".pl","")
       {
        my $path = "$dir/" . $prefix . $file . $suffix;
        return $path if (-r $path && -T $path);
@@ -87,13 +97,13 @@ sub Find
 }
 
 sub findpod {
-    my ($w,$name) = @_;
+    my ($w,$name,%opts) = @_;
+    my $quiet = delete $opts{-quiet};
+    warn "Unhandled extra options: ". join " ", %opts
+	if %opts;
     unless (defined $name and length $name) {
-	$w->messageBox(
-	  -title => "Tk::Pod Error",
-          -message => "Empty POD file/name",
-	);
-	die;
+	return if $quiet;
+	$w->_die_dialog("Empty Pod file/name");
     }
 
     my $absname;
@@ -101,19 +111,15 @@ sub findpod {
 	$absname = $name;
     } else {
 	if ($name !~ /^[-_+:.\/A-Za-z0-9]+$/) {
-	    $w->messageBox(
-	      -title => "Tk::Pod Error",
-	      -message => "Invalid path/file/module name: '$name'\n");
-	    die;
+	    return if $quiet;
+	    $w->_die_dialog("Invalid path/file/module name '$name'\n");
 	}
 	$absname = Find($name);
     }
     if (!defined $absname) {
-	$w->messageBox(
-	  -title => "Tk::Pod Error",
-	  -message => "Can't find POD. Invalid file/module name: '$name'\n"
-	);
-	die;
+	return if $quiet;
+	$w->_error_dialog("Can't find Pod '$name'\n");
+	die "Can't find Pod '$name' in @POD\n";
     }
     if (eval { require File::Spec; File::Spec->can("rel2abs") }) {
 	DEBUG and warn "Turn $absname into an absolute file name";
@@ -122,42 +128,102 @@ sub findpod {
     $absname;
 }
 
+sub _remember_old {
+    my $w = shift;
+    for (qw(File Text)) {
+	$w->{"Old$_"} = $w->{$_};
+    }
+}
+
+sub _restore_old {
+    my $w = shift;
+    for (qw(File Text)) {
+	$w->{$_} = $w->{"Old$_"};
+    }
+}
+
 sub file {   # main entry point
   my $w = shift;
   if (@_)
     {
-      #print "loading $_[0] ...\n";
       my $file = shift;
-      $w->{'File'} = $file;
-      my $path = $w->findpod($file);
-      if (!$w->privateData()->{'from_history'}) {
-	  $w->history_modify_entry;
-	  $w->history_add($path, "1.0");
+      $w->_remember_old;
+      eval {
+	  $w->{'File'} = $file;
+	  $w->{'Text'} = undef;
+	  my $path = $w->findpod($file);
+	  if (!$w->privateData()->{'from_history'}) {
+	      $w->history_modify_entry;
+	      $w->history_add({file => $path}, "1.0");
+	  }
+	  $w->configure('-path' => $path);
+	  $w->delete('1.0' => 'end');
+	  my $tree_sw = $w->parent->Subwidget("tree");
+	  if ($tree_sw) {
+	      $tree_sw->SeePath("file:$path");
+	  }
+	  my $t;
+	  if (DEBUG) {
+	      require Benchmark;
+	      $t = Benchmark->new;
+	  }
+	  if (!$w->get_from_cache) {
+	      $w->process($path);
+	      $w->add_to_cache; # XXX pass time for processing?
+	  }
+	  if (defined $t) {
+	      print Benchmark::timediff(Benchmark->new, $t)->timestr,"\n";
+	  }
+	  $w->focus;
+      };
+      if ($@) {
+	  $w->_restore_old;
+	  die $@;
       }
-      $w->configure('-path' => $path);
-      $w->delete('1.0' => 'end');
-      my $tree_sw = $w->parent->Subwidget("tree");
-      if ($tree_sw) {
-	  $tree_sw->SeePath("file:$path");
-      }
-      my $t;
-      if (DEBUG) {
-	  require Benchmark;
-	  $t = Benchmark->new;
-      }
-      if (!$w->get_from_cache) {
-	  $w->process($path);
-	  $w->add_to_cache; # XXX pass time for processing?
-      }
-      if (defined $t) {
-	  print Benchmark::timediff(Benchmark->new, $t)->timestr,"\n";
-      }
-      $w->focus;
     }
-  else
+  $w->{'File'};
+}
+
+sub text {
+  my $w = shift;
+  if (@_)
     {
-      return $w->{'File'};
+      my $text = shift;
+      $w->_remember_old;
+      eval {
+	  $w->{'Text'} = $text;
+	  $w->{'File'} = undef;
+	  if (!$w->privateData()->{'from_history'}) {
+	      $w->history_modify_entry;
+	      $w->history_add({text => $text}, "1.0");
+	  }
+	  $w->configure('-path' => undef);
+	  $w->delete('1.0' => 'end');
+## XXX Implementation unclear, maybe should be done in showcommand call...
+#	  my $tree_sw = $w->parent->Subwidget("tree");
+#	  if ($tree_sw) {
+#	      $tree_sw->SeeFunc("file:$path");
+#	  }
+	  my $t;
+	  if (DEBUG) {
+	      require Benchmark;
+	      $t = Benchmark->new;
+	  }
+	  # No caching here
+	  # XXX title: the 2nd part of the hack
+	  my $title = $w->cget(-title);
+	  $w->process(\$text, $title);
+	  if (defined $t) {
+	      print Benchmark::timediff(Benchmark->new, $t)->timestr,"\n";
+	  }
+	  $w->focus;
+      };
+      if ($@) {
+	  $w->_restore_old;
+	  die $@;
+      }
     }
+  $w->{'Text'};
 }
 
 sub reload
@@ -166,6 +232,7 @@ sub reload
  # remember old y position
  my ($currpos) = $w->yview;
  $w->delete('0.0','end');
+ $w->delete_from_cache;
  $w->process($w->cget('-path'));
  # restore old y position
  $w->yview(moveto => $currpos);
@@ -173,33 +240,65 @@ sub reload
  $w->markSet(insert => '@0,0');
 }
 
+# Works also for viewing source code
+sub _get_editable_path
+{
+ my ($w) = @_;
+ my $path = $w->cget('-path');
+ if (!defined $path)
+  {
+   my $text = $w->cget("-text");
+   $w->_need_File_Temp;
+   my($fh,$fname) = File::Temp::tempfile(UNLINK => 1,
+					 SUFFIX => ".pod");
+   print $fh $text;
+   close $fh;
+   $path = $fname;
+  }
+ $path;
+}
+
 sub edit
 {
- my ($w,$edit) = @_;
- my $path = $w->cget('-path');
- if ($^O eq 'MSWin32') # XXX what is right?
+ my ($w,$edit,$linenumber) = @_;
+ my $path = $w->_get_editable_path;
+ if (!defined $edit)
   {
-   system("ptked $path");
+   $edit = $ENV{TKPODEDITOR};
+  }
+ if ($^O eq 'MSWin32')
+  {
+   if (defined $edit && $edit ne "")
+    {
+     system(1, $edit, $path);
+    }
+   else
+    {
+     system(1, "ptked", $path);
+    }
   }
  else
   {
-   if (!defined $edit)
-    {
-     $edit = $ENV{TKPODEDITOR};
-    }
-   if (!defined $edit)
+   if (!defined $edit || $edit eq "")
     {
      # VISUAL and EDITOR are supposed to have a terminal, but tkpod can
      # be started without a terminal.
      my $isatty = is_interactive();
-     $edit = $ENV{XEDITOR};
-     if (!$isatty && !defined $edit)
+     if (!$isatty)
       {
-       $w->messageBox(
-	 -title => "Tk::Pod Error",
-         -message => "No terminal, fallback to ptked"
-       );
-       $edit = 'ptked';
+       if (!defined $edit || $edit eq "")
+        {
+         $edit = $ENV{XEDITOR};
+        }
+       if (!defined $edit || $edit eq "")
+        {
+         if (!$terminal_fallback_warn_shown)
+	  {
+           $w->_warn_dialog("No terminal and neither TKPODEDITOR nor XEDITOR environment variables set. Fallback to ptked.");
+	   $terminal_fallback_warn_shown = 1;
+          }
+         $edit = 'ptked';
+        }
       }
      else
       {
@@ -224,46 +323,116 @@ sub edit
        else
         {
          # grandchild
-         exec("$edit $path");
+	 if (defined $linenumber && $edit =~ m{\bemacsclient\b}) # XXX an experiment, maybe support more editors?
+	  {
+	   exec("$edit +$linenumber $path");
+          }
+	 else
+	  {
+           exec("$edit $path");
+          }
         }
       }
     }
   }
 }
 
+sub edit_get_linenumber
+{
+ my($w) = @_;
+ my $linenumber;
+ for my $tag ($w->tagNames('@' . ($w->{MenuX} - $w->rootx) . ',' . ($w->{MenuY} - $w->rooty)))
+  {
+   if ($tag =~ m{start_line_(\d+)})
+    {
+     $linenumber = $1;
+     last;
+    }
+  }
+ $w->edit(undef, $linenumber);
+}
+
+sub view_source
+{
+ my($w) = @_;
+ # XXX why is -title empty here?
+ my $title = $w->cget(-title) || $w->cget('-file');
+ my $t = $w->Toplevel(-title => "Source of $title - Tkpod");
+ my $font_size = $w->base_font_size;
+ my $more = $t->Scrolled('More',
+			 -font => "Courier $font_size",
+			 -scrollbars => $Tk::platform eq 'MSWin32' ? 'e' : 'w',
+			)->pack(-fill => "both", -expand => 1);
+ $more->Load($w->_get_editable_path);
+ $more->AddQuitBindings;
+ $more->focus;
+}
+
+sub copy_pod_location
+{
+ my($w) = @_;
+ my $file = $w->_get_editable_path;
+ if (!defined $file)
+  {
+   $w->_error_dialog("Cannot copy location: this Pod is not associated with a file");
+   return;
+  }
+ $w->SelectionOwn;
+ $w->SelectionHandle(sub {
+			 my($offset,$maxbytes) = @_;
+			 # XXX It's not exactly clear why I have to
+			 # call _get_editable_path again here and not
+			 # reuse $file.
+			 my $file = $w->_get_editable_path;
+			 return undef if $offset > length($file);
+			 substr($file, $offset, $maxbytes);
+		     });
+}
+
+sub _sgn { $_[0] cmp 0 }
+
 sub zoom_normal {
     my $w = shift;
     $w->adjust_font_size($w->standard_font_size);
+    $w->clear_cache;
 }
 
 # XXX should use different increments for different styles
-sub zoom_in {
-    my $w = shift;
-    $w->adjust_font_size($w->base_font_size-1);
-}
-
 sub zoom_out {
     my $w = shift;
-    $w->adjust_font_size($w->base_font_size+1);
+    $w->adjust_font_size($w->base_font_size - 1 * _sgn($w->base_font_size));
+    $w->clear_cache;
 }
 
+sub zoom_in {
+    my $w = shift;
+    $w->adjust_font_size($w->base_font_size + 1 * _sgn($w->base_font_size));
+    $w->clear_cache;
+}
+
+sub More_Widget { "More" }
+sub More_Module { "Tk::More" }
+
 sub Populate
 {
     my ($w,$args) = @_;
 
-    require Tk::More;
+    if ($w->More_Module) {
+	eval q{ require } . $w->More_Module;
+	die $@ if $@;
+    }
 
     $w->SUPER::Populate($args);
 
     $w->privateData()->{history} = [];
     $w->privateData()->{history_index} = -1;
 
-    my $p = $w->Scrolled('More',
+    my $p = $w->Scrolled($w->More_Widget,
 			 -helpcommand => sub {
 			     $w->parent->help if $w->parent->can('help');
 			 },
 			 -scrollbars => $Tk::platform eq 'MSWin32' ? 'e' : 'w');
-    my $p_scr = $p->Subwidget('more');
+    my $p_scr = $p->Subwidget('scrolled');
     $w->Advertise('more' => $p_scr);
     $p->pack(-expand => 1, -fill => 'both');
 
@@ -271,19 +440,15 @@ sub Populate
     #     delegate bind, bindtag to the scrolled widget. Tk402.* (and before?)
     #	  (patch posted and included in Tk402.004)
     $p_scr->bindtags([$p_scr, $p_scr->bindtags]);
-    $p_scr->bind('<Double-1>',       sub  { $w->DoubleClick($_[0]) });#[$w, 'DoubleClick']);
-    $p_scr->bind('<Shift-Double-1>', sub  { $w->ShiftDoubleClick($_[0]) });#[$w, 'ShiftDoubleClick', $_[0]]);
+    $p_scr->bind('<Double-1>',       sub  { $w->DoubleClick($_[0]) });
+    $p_scr->bind('<Shift-Double-1>', sub  { $w->ShiftDoubleClick($_[0]) });
+    $p_scr->bind('<Double-2>',       sub  { $w->ShiftDoubleClick($_[0]) });
+    $p_scr->bind('<3>',              sub  { $w->PostPopupMenu($p_scr, $w->pointerxy) });
+    $p_scr->bind('<2>',		     sub  { $w->OpenPodBySelection });
 
     $p->configure(-font => $w->Font(family => 'courier'));
 
-#XXX del:
-    #$p->tag('configure','verbatim', -wrap => 'none');
     $p->tag('configure','text', -font => $w->Font(family => 'times'));
-    #$p->tag('configure','C',-font => $w->Font(family=>'courier',   weight=>'medium'              ));
-    #$p->tag('configure','B',-font => $w->Font(                     weight=>'bold',               ));
-    ##$p->tag('configure','S',-font => $w->Font(                                                   ));
-    #$p->tag('configure','I',-font => $w->Font(                                       slant => 'i'));
-    #$p->tag('configure','F',-font => $w->Font(                                       slant => 'i'));
 
     $p->insert('0.0',"\n");
 
@@ -302,14 +467,17 @@ sub Populate
     };
 
     my $m = $p->Menu
-	(-tearoff => $Tk::platform ne 'MSWin32',
+	(-title => "Tkpod",
+	 -tearoff => $Tk::platform ne 'MSWin32',
 	 -menuitems =>
 	 [
 	  [Button => 'Back',     -command => [$w, 'history_move', -1]],
 	  [Button => 'Forward',  -command => [$w, 'history_move', +1]],
 	  [Button => 'Reload',   -command => sub{$w->reload} ],
-	  [Button => 'Edit POD',       -command => sub{$w->edit} ],
-	  [Button => 'Search fulltext',-command => ['SearchFullText', $w]],
+	  [Button => 'Edit Pod',       -command => sub{ $w->edit_get_linenumber } ],
+	  [Button => 'View source',    -command => sub{ $w->view_source } ],
+	  [Button => 'Copy Pod location', -command => sub { $w->copy_pod_location } ],
+	  [Button => 'Search full text',-command => ['SearchFullText', $w]],
 	  [Separator => ""],
 	  [Cascade => 'Edit',
 	   ($Tk::VERSION > 800.015 && $p->can('EditMenuItems') ? (-menuitems => $p->EditMenuItems) : ()),
@@ -329,13 +497,16 @@ sub Populate
 
     $w->ConfigSpecs(
             '-file'       => ['METHOD'  ],
+            '-text'       => ['METHOD'  ],
             '-path'       => ['PASSIVE' ],
             '-poddone'    => ['CALLBACK'],
+	    '-title'      => ['PASSIVE' ], # XXX unclear
 
             '-wrap'       => [ $p, qw(wrap       Wrap       word) ],
 	    # -font ignored because it does not change the other fonts
 	    #'-font'	  => [ 'PASSIVE', undef, undef, undef],
             '-scrollbars' => [ $p, qw(scrollbars Scrollbars), $Tk::platform eq 'MSWin32' ? 'e' : 'w' ],
+	    '-basefontsize' => ['METHOD'], # XXX may change
 
             'DEFAULT'     => [ $p ],
             );
@@ -343,7 +514,18 @@ sub Populate
     $args->{-width} = $w->{Length};
 }
 
-#my %tag = qw(C 1 B 1 I 1 L 1 F 1 S 1 Z 1); # un-used XXX
+sub basefontsize
+{
+ my($w, $val) = @_;
+ if ($val)
+  {
+   $w->set_base_font_size($val);
+  } 
+ else
+  {
+   $w->base_font_size;
+  }
+}
 
 sub Font
 {
@@ -374,18 +556,18 @@ sub DoubleClick
    if ($file = $w->findpod($sel)) {
        if (defined $how && $how eq 'new')
 	{
-         $w->MainWindow->Pod('-file' => $sel);
+	 my $tree = eval { $w->parent->cget(-tree) };
+	 my $exitbutton = eval { $w->parent->cget(-exitbutton) };
+         $w->MainWindow->Pod('-file' => $sel,
+			     '-tree' => $tree,
+			     -exitbutton => $exitbutton);
 	}
        else
 	{
          $w->configure('-file'=>$file);
         }
    } else {
-       $w->messageBox(
-         -title => "Tk::Pod Error",
-         -message => "No Pod documentation found for '$sel'\n"
-       );
-       die;
+       $w->_die_dialog("No Pod documentation found for '$sel'\n");
    }
   }
  Tk->break;
@@ -406,19 +588,30 @@ sub Link
   {
    my $file = $w->cget('-file');
    $w->configure('-file' => $man)
-    unless ( defined $file and ($file =~ /$man\.\w+$/ or $file eq $man) );
+    unless ( defined $file and ($file =~ /\Q$man\E\.\w+$/ or $file eq $man) );
   }
 
  if ($how eq 'new')
   {
    $man = $w->cget('-file') if ($man eq "");
    my $tree = eval { $w->parent->cget(-tree) };
-   $w = $w->MainWindow->Pod('-tree' => $tree);
-   $w->configure('-file' => $man); # see tkpod for the same problem
+   my $exitbutton = eval { $w->parent->cget(-exitbutton) };
+   my $old_w = $w;
+   my $new_pod = $w->MainWindow->Pod('-tree' => $tree,
+				     -exitbutton => $exitbutton,
+				    );
+   $new_pod->configure('-file' => $man); # see tkpod for the same problem
+
+   $w = $new_pod->Subwidget('pod');
+   # set search term for new window
+   my $search_term_ref = $old_w->Subwidget('more')->Subwidget('searchentry')->cget(-textvariable);
+   if (defined $$search_term_ref && $$search_term_ref ne "") {
+       $ {$w->Subwidget('more')->Subwidget('searchentry')->cget(-textvariable) } = $$search_term_ref;
+   }
   }
   # XXX big docs like Tk::Text take too long until they return
 
- if ($sec ne '' && $man eq '') # XXX reuse vs. new`
+ if ($sec ne '' && $man eq '') # XXX reuse vs. new
   {
    $w->history_modify_entry;
   }
@@ -477,23 +670,133 @@ sub Link
     {
      DEBUG and warn " Not found! (\"sec\")\n";
 
-     $w->messageBox(
-       -title   => "Tk::Pod Error",
-       -message => "Section '$sec' not found\n"
-     );
-     die;
+     $w->_die_dialog("Section '$sec' not found\n");
     }
    DEBUG and warn "link-zapping to $start linestart\n";
    $w->yview("$start linestart");
   }
 
- if ($sec ne '' && $man eq '') # XXX reuse vs. new`
+ if ($sec ne '' && $man eq '') # XXX reuse vs. new
   {
-   $w->history_add($w->cget(-path), $w->index('@0,0'));
+   $w->history_add({file => $w->cget(-path)}, $w->index('@0,0'));
   }
 
 }
 
+sub Link_url {
+    my ($w,$how,$index,$man,$sec) = @_;
+    DEBUG and warn "Start browser with $man\n";
+    start_browser($man);
+}
+
+sub Link_man {
+    my ($w,$how,$index,$man,$sec) = @_;
+    my $mansec;
+    if ($man =~ s/\s*\((.*)\)\s*$//) {
+	$mansec = $1;
+    }
+    my @manbrowser;
+    if (exists $ENV{TKPODMANVIEWER} && $ENV{TKPODMANVIEWER} eq "internal") {
+	DEBUG and warn "Use internal man viewer\n";
+    } else {
+	my $manurl = "man:$man($mansec)";
+	if (defined $sec && $sec ne "") {
+	    $manurl .= "#$sec";
+	}
+	DEBUG and warn "Try to start any man browser for $manurl\n";
+	@manbrowser = ('gnome-help-browser', 'khelpcenter');
+	my $wm = detect_window_manager($w);
+	DEBUG and warn "Window manager system is $wm\n";
+	if ($wm eq 'kde') {
+	    unshift @manbrowser, 'khelpcenter';
+	}
+	if (defined $ENV{TKPODMANVIEWER}) {
+	    unshift @manbrowser, $ENV{TKPODMANVIEWER};
+	}
+	for my $manbrowser (@manbrowser) {
+	    DEBUG and warn "Try $manbrowser...\n";
+	    if (is_in_path($manbrowser)) {
+		if (fork == 0) {
+		    DEBUG and warn "Use $manbrowser...\n";
+		    exec($manbrowser, $manurl);
+		    die $!;
+		}
+		return;
+	    }
+	}
+    }
+    if (!$w->InternalManViewer($mansec, $man)) {
+	$w->_die_dialog("No useable man browser found. Tried @manbrowser and internal man viewer via `man'");
+    }
+}
+
+sub InternalManViewer {
+    my($w, $mansec, $man) = @_;
+    my $man_exe = "man";
+    if (!is_in_path($man_exe)) {
+	if ($^O eq 'MSWin32') {
+	    $man_exe = "c:/cygwin/bin/man.exe";
+	    if (!-e $man_exe) {
+		return 0;
+	    }
+	} else {
+	    return 0;
+	}
+    }
+    my $t = $w->Toplevel(-title => "Manpage $man($mansec)");
+    my $font_size = $w->base_font_size;
+    my $more = $t->Scrolled("More",
+			    -font => "Courier $font_size",
+			    -scrollbars => $Tk::platform eq 'MSWin32' ? 'e' : 'w',
+			   )->pack(-fill => "both", -expand => 1);
+    $more->tagConfigure("bold", -font => "Courier $font_size bold");
+    my $menu = $more->menu;
+    $t->configure(-menu => $menu);
+    local $SIG{PIPE} = "IGNORE";
+    my $can_langinfo = $] >= 5.008 && eval { require I18N::Langinfo; 1 };
+    local $ENV{LANG} = $ENV{LANG};
+    if (!$can_langinfo) {
+	$ENV{LANG} = "C";
+    }
+    open(MAN, $man_exe . (defined $mansec ? " $mansec" : "") . " $man |")
+	or die $!;
+    if ($can_langinfo) {
+	my $codeset = I18N::Langinfo::langinfo(I18N::Langinfo::CODESET());
+	eval qq{ binmode MAN, q{:encoding($codeset)} };
+	warn $@ if $@;
+    }
+    if (eof MAN) {
+	$more->insert("end", "No entry for for $man" . (defined $mansec ? " in section $mansec of" : "") . " the manual");
+    } else {
+	while(<MAN>) {
+	    chomp;
+	    (my $line = $_) =~ s/.\cH//g;
+	    my @bold;
+	    while (/(.*?)((?:(.)(\cH\3)+)+)/g) {
+		my($pre, $bm) = ($1, $2);
+		$pre =~ s/.\cH//g;
+		$bm  =~ s/.\cH//g;
+		push @bold, length $pre, length $bm;
+	    }
+	    if (@bold) {
+		my $is_bold = 0;
+		foreach my $length (@bold) {
+		    if ($length > 0) {
+			(my($s), $line) = $line =~ /^(.{$length})(.*)/;
+			$more->insert("end", $s, $is_bold ? "bold" : ());
+		    }
+		    $is_bold = 1 - $is_bold;
+		}
+		$more->insert("end", "$line\n");
+	    } else {
+		$more->insert("end", "$line\n");
+	    }
+	}
+    }
+    close MAN;
+    1;
+}
+
 sub EnterLink {
     my $w = shift;
     $w->configure(-cursor=>'hand2');
@@ -509,6 +812,14 @@ sub SearchFullText {
     unless (defined $IDX && $IDX->IsWidget) {
 	require Tk::Pod::Search; #
 	$IDX = $w->Toplevel(-title=>'Perl Library Full Text Search');
+	$IDX->transient($w);
+
+	my $current_path;
+	my $tree_sw = $w->parent->Subwidget("tree");
+	if ($tree_sw) {
+	    $current_path = $tree_sw->GetCurrentPodPath;
+	}
+
 	$IDX->PodSearch(
 			-command =>
 			sub {
@@ -522,8 +833,12 @@ sub SearchFullText {
 				 -searchterm => $args{-searchterm},
 				 -onlymatch => 1,
 				);
-			}
+			},
+			-currentpath => $current_path,
 		       )->pack(-fill=>'both',-expand=>'both');
+	$IDX->Button(-text => "Close",
+		     -command => sub { $IDX->destroy },
+		    )->pack(-fill => 'x');
     }
     $IDX->deiconify;
     $IDX->raise;
@@ -531,15 +846,30 @@ sub SearchFullText {
     (($IDX->children)[0])->focus;
 }
 
+sub _need_File_Temp {
+    my $w = shift;
+    if (!eval { require File::Temp; 1 }) {
+	$w->_die_dialog("The perl module 'File::Temp' is missing");
+    }
+}
+
 sub Print {
     my $w = shift;
-    my $path = $w->cget(-path);
-    if (!-r $path) {
-	$w->messageBox(
-          -title   => "Tk::Pod Error",
-	  -message => "Cannot find file `$path`"
-	);
-	die;
+
+    my($text, $path);
+    $path = $w->cget(-path);
+    if (defined $path) {
+	if (!-r $path) {
+	    $w->_die_dialog("Cannot find file `$path`");
+	}
+    } else {
+	$text = $w->cget("-text");
+	$w->_need_File_Temp;
+	my($fh,$fname) = File::Temp::tempfile(UNLINK => 1,
+					      SUFFIX => ".pod");
+	print $fh $text;
+	close $fh;
+	$path = $fname;
     }
 
     if ($ENV{'TKPODPRINT'}) {
@@ -552,30 +882,53 @@ sub Print {
     }
     # otherwise fall thru...
 
-    if (!eval { require POSIX; 1 }) {
-	$w->messageBox(
-          -title   => "Tk::Pod Error",
-	  -message => "The perl module 'POSIX' is missing"
-	);
-	die;
+    my $success = $w->_print_pod_unix($path);
+
+    if (!$success) {
+	$w->_error_dialog("Can't print on your system.\nEither pod2man, groff,\ngv or ghostview are missing.");
     }
+}
+
+sub _print_pod_unix {
+    my($w, $path) = @_;
     if (is_in_path("pod2man") && is_in_path("groff")) {
-	my $gv = is_in_path("gv") || is_in_path("ghostview") || is_in_path("XXXggv") || is_in_path("kghostview");
+	my $pod2ps_pipe = "pod2man $path | groff -man -Tps";
+
+	if ($^O eq 'darwin') {
+	    my $cmd = "$pod2ps_pipe | /usr/bin/open -a Preview -f";
+	    system($cmd) == 0
+		or $w->_die_dialog("Error while executing <$cmd>. Status code is $?");
+	    return 1;
+	}
+
+	# XXX maybe determine user's environment (GNOME vs. KDE vs. plain X11)?
+	my $gv = is_in_path("gv")
+	      || is_in_path("ghostview")
+	      || is_in_path("ggv")         # newer versions seem to work
+	      || is_in_path("kghostview");
 	if ($gv) {
-	    my $temp = POSIX::tmpnam();
-	    # XXX $temp is never deleted...
-	    system("pod2man $path | groff -man -Tps > $temp");
-	    system("$gv $temp &");
-	    return;
+	    $w->_need_File_Temp;
+
+	    my($fh,$fname) = File::Temp::tempfile(SUFFIX => ".ps");
+	    system("$pod2ps_pipe > $fname");
+	    push @tempfiles, $fname;
+	    my $pid = fork;
+	    if (!defined $pid) {
+		die "Can't fork: $!";
+	    }
+	    if ($pid == 0) {
+		exec($gv, $fname);
+		warn "Exec of $gv $fname failed: $!";
+		CORE::exit(1);
+	    }
+	    push @gv_pids, $pid;
+	    return 1;
 	}
     }
-    $w->messageBox(
-      -title   => "Tk::Pod Error",
-      -message => "Can't print on your system.\nEither pod2man, groff,\ngv or ghostview are missing."
-    );
-    die;
+    return 0;
 }
 
+
 sub _substitute_cmd {
     my($cmd, $path) = @_;
     my @cmd;
@@ -595,12 +948,13 @@ sub Print_MSWin {
    eval {require Win32; 1} and
    defined(&Win32::GetOSName) and
     (Win32::GetOSName() eq 'Win32s'  or   Win32::GetOSName() eq 'Win95');
-  require POSIX;
+  require POSIX; # XXX should be probably replaced by File::Temp, but I have no Win machine to test...
 
   my $temp = POSIX::tmpnam(); # XXX it never gets deleted
   $temp =~ tr{/}{\\};
   $temp =~ s/\.$//;
   DEBUG and warn "Using $temp as the temp file for hardcopying\n";
+  # XXX cleanup of temp file?
 
   if($is_old) { # so we can't assume that write.exe can handle RTF
     require Pod::Simple::Text;
@@ -620,6 +974,7 @@ sub Print_MSWin {
   return;
 }
 
+sub PrintHasDialog { $^O ne 'MSWin32' }
 
 # Return $first and $last indices of the word under $index
 sub _word_under_index {
@@ -649,18 +1004,29 @@ sub SelectToModule {
 # Add the file $file (with optional text index position $index) to the
 # history.
 sub history_add {
-    my ($w,$file,$index) = @_;
-    unless (-f $file) {
-	$w->messageBox(-message => "Not a file '$file'. Can't add to history\n",
-		       @{&HISTORY_DIALOG_ARGS});
-	return;
+    my ($w,$what,$index) = @_;
+    my($file, $text);
+    if (ref $what eq 'HASH') {
+	$file = $what->{file};
+	$text = $what->{text};
+    } else {
+	$file = $what;
+	$what = {file => $file};
+    }
+    if (defined $file) {
+	unless (-f $file) {
+	    $w->messageBox(-message => "Not a file '$file'. Can't add to history\n",
+			   @{&HISTORY_DIALOG_ARGS});
+	    return;
+	}
     }
     my $hist = $w->privateData()->{history};
-    my $hist_entry = _HistoryEntry->create($file, $index);
+    my $hist_entry = _HistoryEntry->create($what, $index);
     $hist->[++$w->privateData()->{history_index}] = $hist_entry;
     splice @$hist, $w->privateData()->{history_index}+1;
     $w->history_view_update;
     $w->history_view_select;
+    $w->_history_navigation_update;
     undef;
 }
 
@@ -675,7 +1041,7 @@ sub history_back {
 	return;
     }
     if ($w->privateData()->{history_index} <= 0) {
-	$w->messageBox(-message => "Can't go back",
+	$w->messageBox(-message => "Can't go back in history",
 		       @{&HISTORY_DIALOG_ARGS});
 	return;
     }
@@ -696,7 +1062,7 @@ sub history_forward {
 	return;
     }
     if ($w->privateData()->{history_index} >= $#$hist) {
-	$w->messageBox(-message => "Can't go forward",
+	$w->messageBox(-message => "Can't go forward in history",
 		       @{&HISTORY_DIALOG_ARGS});
 	return;
     }
@@ -711,16 +1077,44 @@ sub history_forward {
 sub _history_update {
     my($w, $hist_entry) = @_;
     if ($hist_entry) {
-	if ($w->cget('-path') ne $hist_entry->file) {
+	if (defined $hist_entry->file) {
+	    if ($w->cget('-path') ne $hist_entry->file) {
+		$w->privateData()->{'from_history'} = 1;
+		$w->configure('-file' => $hist_entry->file);
+		$w->privateData()->{'from_history'} = 0;
+	    }
+	} elsif (defined $hist_entry->text) {
 	    $w->privateData()->{'from_history'} = 1;
-	    $w->configure('-file' => $hist_entry->file);
+	    $w->configure('-text' => $hist_entry->text);
 	    $w->privateData()->{'from_history'} = 0;
 	}
+	$w->_history_navigation_update;
 	$w->afterIdle(sub { $w->see($hist_entry->index) })
 	    if $hist_entry->index;
     }
 }
 
+sub _history_navigation_update {
+    my $w = shift;
+    # XXX Be careful with the search pattern
+    # if I decide to I18N Tk::Pod one day...
+    my $m_history;
+    if ($w->parent and $m_history = $w->parent->Subwidget("menubar")) {
+	$m_history = $m_history->entrycget("History", "-menu");
+	my $inx = $w->privateData()->{history_index};
+	if ($inx == 0) {
+	    $m_history->entryconfigure("Back", -state => "disabled");
+	} else {
+	    $m_history->entryconfigure("Back", -state => "normal");
+	}
+	if ($inx == $#{$w->privateData()->{history}}) {
+	    $m_history->entryconfigure("Forward", -state => "disabled");
+	} else {
+	    $m_history->entryconfigure("Forward", -state => "normal");
+	}
+    }
+}
+
 # Move the history backward ($inc == -1) or forward ($inc == +1)
 sub history_move {
     my($w, $inc) = @_;
@@ -743,8 +1137,9 @@ sub history_set {
 sub history_modify_entry {
     my $w = shift;
     if ($w->privateData()->{'history_index'} >= 0) {
-	my $old_entry = _HistoryEntry->create($w->cget('-path'),
-					      $w->index('@0,0'));
+	my $old_entry = _HistoryEntry->create({file => $w->cget('-path'),
+					       text => $w->cget('-text')
+					      }, $w->index('@0,0'));
 	$w->privateData()->{'history'}->[$w->privateData()->{'history_index'}] = $old_entry;
     }
 }
@@ -770,10 +1165,14 @@ sub history_view {
 		      return if !defined $sel;
 		      $w->history_set($sel);
 		  });
+	$t->Button(-text => "Close",
+		   -command => sub { $t->destroy },
+		  )->pack(-fill => 'x');
     }
     $t->deiconify;
     $t->raise;
     $w->history_view_update;
+    $w->history_view_select;
 }
 
 # Re-fill the history view with the current history array.
@@ -801,65 +1200,84 @@ sub history_view_select {
     }
 }
 
-#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+sub PostPopupMenu {
+    my($w, $p_scr, $X, $Y) = @_;
+    $w->{MenuX} = $X;
+    $w->{MenuY} = $Y;
+    $p_scr->PostPopupMenu($X, $Y);
+}
 
-# REPO BEGIN
-# REPO NAME is_in_path /home/e/eserte/src/repository
-# REPO MD5 1b42243230d92021e6c361e37c9771d1
+sub OpenPodBySelection {
+    my($w) = @_;
+    my $sel;
+    Tk::catch {
+	$sel = $w->SelectionGet('-selection' => ($Tk::platform eq 'MSWin32'
+						 ? "CLIPBOARD"
+						 : "PRIMARY"));
+    };
+    $sel =~ s{\s}{}g; # no whitespace in Pod names possible
+    $w->configure(-file => $sel);
+}
 
-sub is_in_path {
-    my($prog) = @_;
-    require Config;
-    my $sep = $Config::Config{'path_sep'} || ':';
-    foreach (split(/$sep/o, $ENV{PATH})) {
-	if ($^O eq 'MSWin32') {
-	    return "$_\\$prog"
-		if (-x "$_\\$prog.bat" ||
-		    -x "$_\\$prog.com" ||
-		    -x "$_\\$prog.exe");
+sub _die_dialog {
+    shift->_error_dialog(@_);
+    die;
+}
+
+sub _error_dialog {
+    my($w, $message) = @_;
+    $w->messageBox(
+      -title   => "Tk::Pod Error",
+      -message => $message,
+      -icon => 'error',
+    );
+}
+
+sub _warn_dialog {
+    my($w, $message) = @_;
+    $w->messageBox(
+      -title   => "Tk::Pod Warning",
+      -message => $message,
+      -icon => 'warning',
+    );
+}
+
+END {
+    if (@tempfiles) {
+	my $gv_running;
+	for my $pid (@gv_pids) {
+	    if (kill 0 => $pid) {
+		$gv_running = 1;
+		last;
+	    }
+	}
+
+	if ($gv_running) {
+	    warn "A ghostscript (or equivalent) process is still running, won't delete temporary files: @tempfiles\n";
 	} else {
-	    return "$_/$prog" if (-x "$_/$prog");
+	    for my $temp (@tempfiles) {
+		unlink $temp;
+	    }
 	}
     }
-    undef;
-}
-# REPO END
-
-sub is_interactive {
-    if ($^O eq 'MSWin32' || !eval { require POSIX; 1 }) {
-	# fallback
-	return -t STDIN && -t STDOUT;
-    }
-
-    # from perlfaq8
-    open(TTY, "/dev/tty") or die $!;
-    my $tpgrp = POSIX::tcgetpgrp(fileno(*TTY));
-    my $pgrp = getpgrp();
-    if ($tpgrp == $pgrp) {
-	1;
-    } else {
-	0;
-    }
 }
 
-#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 1;
 
 __END__
 
 =head1 NAME
 
-Tk::Pod::Text - POD browser widget
-
+Tk::Pod::Text - Pod browser widget
 
 =head1 SYNOPSIS
 
     use Tk::Pod::Text;
 
-    $pod = $parent->PodText(
-		-file		=> ?
-		-scrollbars	=> ?
-		);
+    $pod = $parent->Scrolled("PodText",
+			     -file	 => $file,
+			     -scrollbars => "osoe",
+		            );
 
     $file = $pod->cget('-path');   # ?? the name path is confusing :-(
 
@@ -872,9 +1290,43 @@ Tk::Pod::Text - POD browser widget
 
 =head1 DESCRIPTION
 
-B<Tk::Pod::Text> is a readonly text widget that can display POD
+B<Tk::Pod::Text> is a readonly text widget that can display Pod
 documentation.
 
+=head1 OPTIONS
+
+=over
+
+=item -file
+
+The named (pod) file to be displayed.
+
+=item -path
+
+Return the expanded path of the currently displayed Pod. Useable only
+with the C<cget> method.
+
+=item -poddone
+
+A callback to be called if parsing and displaying of the Pod is done.
+
+=item -wrap
+
+Set the wrap mode. Default is C<word>.
+
+=item -scrollbars
+
+The position of the scrollbars, see also L<Tk::Scrolled>. By default,
+the vertical scrollbar is on the right on Windows systems and on the
+left on X11 systems.
+
+Note that it is not necessary and usually will do the wrong thing if
+you put a C<Tk::Pod::Text> widget into a C<Scrolled> component.
+
+=back
+
+Other options are propagated to the embedded L<Tk::More> widget.
+
 =head1 ENVIRONMENT
 
 =over
@@ -887,8 +1339,8 @@ Turn debugging mode on if set to a true value.
 
 Use the specified program for printing the current pod. If the string
 contains a C<%s>, then filename substitution is used, otherwise the
-filename of the POD document is appended to the value of
-C<TKPODPRINT>. Here is a silly example to send the POD to a web browser:
+filename of the Pod document is appended to the value of
+C<TKPODPRINT>. Here is a silly example to send the Pod to a web browser:
 
     env TKPODPRINT="pod2html %s > %s.html; galeon %s.html" tkpod ...
 
@@ -900,6 +1352,15 @@ C<XEDITOR>, C<VISUAL>, or C<EDITOR> is used on Unix. As a last
 fallback, C<ptked> or C<vi> are used, depending on platform and
 existance of a terminal.
 
+=item TKPODMANVIEWER
+
+Use the specified program as the manpage viewer. The manpage viewer
+should accept a manpage URL (C<man://>I<manpage>(I<section>)).
+Alternatively the special viewer "internal" may be used. As fallback,
+the default GNOME and/or KDE manpage viewer will be called.
+
+=back
+
 =head1 SEE ALSO
 
 L<Tk::More|Tk::More>
@@ -915,7 +1376,7 @@ L<perlindex|perlindex>
 
 =head1 KNOWN BUGS
 
-See TODO files of Tk-Pod distribution
+See L<TODO> file of Tk-Pod distribution
 
 
 
@@ -937,7 +1398,11 @@ A file: F</usr/local/bin/perl>.  A variable $a without markup.
 
 S<boofar> is in SE<lt>E<gt>.
 
-German Umlaute:
+Indexed items are not supported in Tk::Pod. X<Index Test>
+
+Zero-Z<>effect formatting.
+
+German umlauts:
 
 =over 4
 
@@ -957,10 +1422,14 @@ German Umlaute:
 
 =back
 
-Pod with Umlaut: L<ExtUtils::MakeMaker> and ExtUtils::MakeMaker.
+Unicode outside Latin1 range: E<0x20ac> (euro sign).
+
+Pod with umlaut: L<ExtUtils::MakeMaker>.
 
 Details:  L<perlpod> or perl, perlfunc.
 
+External links: L<http://www.cpan.org> (URL), L<perl(1)> (man page).
+
 Here some code in a as is paragraph
 
     use Tk;
@@ -977,13 +1446,67 @@ Mixed Fonts: B<C<bold-fixed>>, B<I<bold-italics>>
 
 Non-breakable text: S<The quick brown fox jumps over the lazy fox.>
 
-Other POD docu: Tk::Font, Tk::BrowseEntry
+Modern Pod constructs (multiple E<lt>E<gt>): I<< italic >>, C<< fixed
+with embedded < and > >>.
+
+Itemize with numbers:
+
+=over
+
+=item 1.
+
+First
+
+=item 2.
+
+Second
+
+=item 3.
+
+Thirs
+
+=back
+
+Itemize with bullets:
+
+=over
+
+=item *
+
+First
+
+=item *
+
+Second
+
+=item *
+
+Thirs
+
+=back
+
+=head1 TESTING HEAD1
+
+=head2 TESTING HEAD2
+
+=head3 TESTING HEAD3
+
+=head4 TESTING HEAD4
+
+=begin a_format_which_does_not_exist
+
+This section should be invisible (=begin and =end).
+
+=end a_format_which_does_not_exist
+
+Other Pod docu: Tk::Font, Tk::BrowseEntry (not underlined, but
+double-clickable in Tk::Pod)
 
 =head1 AUTHOR
 
 Nick Ing-Simmons <F<nick@ni-s.u-net.com>>
 
-Current maintainer is Slaven Rezic <F<slaven.rezic@berlin.de>>.
+Current maintainer is Slaven ReziE<0x107> <F<slaven@rezic.de>>.
 
 Copyright (c) 1998 Nick Ing-Simmons.  All rights reserved.  This program
 is free software; you can redistribute it and/or modify it under the same
